--------------------------------------------------------
-- D2IO_driver Module
--
-- FILE NAME:		D2IO_driver.vhd
-- ORIG AUTHOR:	EE480
-- COMPANY:			University of Alberta
--	DATE:				July, 2002
--
--	DESCRIPTION:	
--
--			Interface module to the D2IO board.
--			Interacts with CPLD on D2IO board to
--			control the following I/O:
--
--			7Seg Display 	(2)		(write)
--			LED				(16)		(write)
--			2Pos Switch		(8)		(read)
--			PushButtons		(15)		(read)
--
--			All I/O is made available outside this module 
--			via a set of registers. The registers are described
--			as follows:
--
--			RegNAME		Size		Read/Write (R/W)
--			-------		----		----------------
--			pb_reg		15			Read Only
--			sw_reg		8			Read Only
--			led_signal	16			Write Only
--			sseg_signal	16			Write Only	
--							
--			The registers correspond to the physical hardware 
--			in the following manner:
--
--			PB_REG(14 downto 0):
--			------------------------
--			PB_REG(n)	-> 	B(n)
--
--			The puchbuttons read as ACTIVE HIGH. 
--			This means in their normal state, they 
--			read as '0', and as '1' when pressed.
--
--			SW_REG(7 downto 0):
--			------------------------
--			SW_REG(n)	->		SW(n)
--
--			Some experimentation may be necessary 
--			to determine the logical values of the 
--			switch positions. There are different
--			versions of the D2IO board, some are
--			'active' (although there really is no active 
--			state) HIGH, and some 'active' LOW.
--
--			D2IO rev.C will detect a logic '1' when 
--			the switches are 'ON'. see diagram.
--
--					ON			OFF
--					___		___
--				  | 0 |	  |   |
--				  |   |    |   |
--				  |___|    |_0_|
--
--
--			LED_SIGNAL(15 downto 0):
--			------------------------
--			LED_SIGNAL(n)	->	LD(n)
--
--			The LED's are active high. Write a '1'
--			to turn on an LED.
--
--			SSEG_SIGNAL(15 downto 0):
--			-------------------------
--			SSEG_SIGNAL(15 downto 12)	->	Seg1	
--			SSEG_SIGNAL(11 downto 8)	->	Seg2
--			SSEG_SIGNAL(7 downto 4)		->	Seg3
--			SSEG_SIGNAL(3 downto 0)		->	Seg4
--
--			Where:
--			
--			Seg1	Seg2	Seg3 	Seg4
--			 __	 __    __    __   
--			|  | 	|  |  |  |  |  |  
--			 --	 --	 --    --	
--			|__|  |__|  |__|  |__|
--		 
--			Data written to the SSEG register will be 
--			displayed as in a hexidecimal representations.
--			At this point there is NO direct control over
--			the individual segments as the low level segment
--			drivers are contained in the NV-CPLD.
--------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity d2IO_driver is
	generic( AWidth : integer := 6;
				DWidth : integer := 8);
    Port ( 
	 			clk : in std_logic;										--sytem clock (50MHz Xtal)
				IO_clk: out std_logic;									--output clock (divided)
	         
				pb_reg : out std_logic_vector (14 downto 0); 	--register for external pushbutton reads
				sw_reg : out std_logic_vector (7 downto 0);		--register for external 2pos switches
				led_signal : in std_logic_vector (15 downto 0);	--register to write to LEDs
				sseg_signal : in std_logic_vector (15 downto 0);--register to write to 7-Segment Display	
			
				cs : out std_logic;										--chip select
           	we : out std_logic;										--write enable
           	oe : out std_logic;										--output enable
 				addr : out std_logic_vector(5 downto 0);			--address bus for communications with CPLD
           	data : inout std_logic_vector(7 downto 0)			--bidir data bus between CPLD & design
			);
end d2IO_driver;

architecture Behavioral of d2IO_driver is
	
	constant SSL_adr : std_logic_vector(AWidth - 1 downto 0) := "000111";
	constant SSH_adr : std_logic_vector(AWidth - 1 downto 0) := "000110";
	constant SWT_adr : std_logic_vector(AWidth - 1 downto 0) := "000011";
	constant PBL_adr : std_logic_vector(AWidth - 1 downto 0) := "000000";
	constant PBH_adr : std_logic_vector(AWidth - 1 downto 0) := "000001";
	constant LEDL_adr : std_logic_vector(AWidth - 1 downto 0) := "000100";
	constant LEDH_adr : std_logic_vector(AWidth - 1 downto 0) := "000101";
	
	component clock_divider
		generic (Divider1, Divider2 : integer);
		port (
			Clk_in : in std_logic;
			Clk_out : out std_logic
		  		);
	end component;
	
	type state_type is ( waiting, led_strobe_1, led_latch_1, led_strobe_2, 
								led_latch_2, sseg_strobe_1, sseg_latch_1, sseg_strobe_2, 
								sseg_latch_2, sw_read, pb_read_7_downto_0, pb_read_15_downto_8,
								update, polling
								);

	signal clk_slow : std_logic;
	signal pb_signal : std_logic_vector(15 downto 0);
	signal led_reg : std_logic_vector (15 downto 0);
	signal sseg_reg : std_logic_vector (15 downto 0);
	signal d2_state, d2_next_state : state_type;

begin

	cs <= '1';									-- chip select for the fpga is always on
	pb_reg <= pb_signal(14 downto 0);	--disregard MSB as there are only 15 btns
	IO_clk<= clk_slow;

--instantiate clock divider
DivClk : component clock_divider
			generic map (	Divider1 =>8,
		 						Divider2 =>1)
			port map (	Clk_in => clk,
		 					Clk_out => clk_slow); 

	--next state transition
	process (clk_slow)
	begin
		if rising_edge(clk_slow) then
			d2_state <= d2_next_state;
		end if;
	end process;

	--update data inputs on change
	--reorder the 7Seg bits so they can be written to in a logical orger
	process (led_signal, sseg_signal)
	begin
		led_reg <= led_signal;		
		sseg_reg(15 downto 12) <= sseg_signal(11 downto 8);	 
		sseg_reg(11 downto 8) <= sseg_signal(15 downto 12);	
		sseg_reg(7 downto 4) <= sseg_signal(3 downto 0);	
		sseg_reg(3 downto 0) <= sseg_signal(7 downto 4);	
	end process;

	--Handle all of the sequential READ/WRITE operations
	--continously READ the status of the PSHBUTTONS and SWITCHES
	--continously WRITE to the LED's and 7SEG display's
	process (d2_state)
	begin
		case d2_state is

			when polling =>
				data <= (others => 'Z');
				addr <= (others => '0');
				we <= '0';
				oe <= '1';
				d2_next_state <= waiting;
		
			when waiting =>
				data <= (others => 'Z');
				addr <= (others => '0');
				we <= '0';
				oe <= '0';
				d2_next_state <= sw_read;
					
			--READ SWITCHES
			when sw_read =>								--read all switches
				data <= (others => 'Z');
				addr <= SWT_adr;
				we <= '0';
				oe <= '1';
				sw_reg <= data;		
				d2_next_state <= pb_read_7_downto_0;
			
			--READ PUSH BUTTONS
			when pb_read_7_downto_0 =>					--read first 8 buttons
				data <= (others => 'Z');
				addr <= PBL_adr;
				we <= '0';
				oe <= '1';
				pb_signal(7 downto 0) <= data;	
				d2_next_state <= pb_read_15_downto_8;
			when pb_read_15_downto_8 =>				--read final 7 buttons
				data <= (others => 'Z');				--use 8 bit reg to match databus size
				addr <= PBH_adr;							--output as 15 bits
				we <= '0';
				oe <= '1';
				pb_signal(15 downto 8) <= data;	
				d2_next_state <= led_strobe_1;

			--LED STUFF
			when led_strobe_1 =>							--write LSW LED data to DataBus
				data <= led_reg (7 downto 0);
				addr <= LEDL_adr;
				we <= '1';
				oe <= '0';
				d2_next_state <=led_latch_1;
			when led_latch_1 =>							--latch data on LSW LEDs
				data <= led_reg (7 downto 0);
				addr <= LEDL_adr;
				we <= '0';
				oe <= '0';
				d2_next_state <= led_strobe_2;
			when led_strobe_2 =>							--write MSW LED data to databus
				data <= led_reg (15 downto 8);
				addr <= LEDH_adr;
				we <= '1';
				oe <= '0';
				d2_next_state <= led_latch_2;
			when led_latch_2 =>							--latch data on LSW LEDs
				data <= led_reg (15 downto 8);
				addr <= LEDH_adr;
				we <= '0';
				oe <= '0';
				d2_next_state <= sseg_strobe_1;
			
			--7SEGMENT STUFF
			when sseg_strobe_1 =>						--write to seg3,4
				data <= sseg_reg (7 downto 0);
				addr <= SSL_adr;
				we <= '1';
				oe <= '0';
				d2_next_state <= sseg_latch_1;
			when sseg_latch_1 =>
				data <= sseg_reg (7 downto 0);
				addr <= SSL_adr;
				we <= '0';
				oe <= '0';
				d2_next_state <= sseg_strobe_2;
			when sseg_strobe_2 =>					--write to seg1,2
				data <= sseg_reg (15 downto 8);
				addr <= SSH_adr;
				we <= '1';
				oe <= '0';
				d2_next_state <= sseg_latch_2;
			when sseg_latch_2 =>
				data <= sseg_reg (15 downto 8);
				addr <= SSH_adr;
				we <= '0';
				oe <= '0';
				d2_next_state <= polling;

			when others =>
				d2_next_state <= polling;				--default starting point
			end case;

	end process;

end Behavioral;
