-- rs232wb.vhd
--
-- Author: Tyler Brandon
--
-- Created: Oct 15, 2004
--
-- Desc:  Connects the uart code provided by Xilinx (Ken Chapman) to
--        a wishbone bus.  This module acts as the wishbone master.
-- 

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity rs232wb is
    generic (  
        ADDR_WIDTH : integer := 16;
        DATA_WIDTH : integer := 16
    );
                
    port (
    clk_i                   : IN std_logic;	-- System clock input			   
    rst_i                 : IN std_logic;   	-- Reset signal for this module
    ack_i                   : IN std_logic;   	-- Ack input from Wishbone "slaves"
    err_i                   : IN std_logic;   	-- Err input from Wishbone "slaves"
    master_bg_i             : IN std_logic;   	-- Bus Grant (grants this module the bus)
                             					-- Address bus input from "normal" Wishbone
                             					-- master (i.e. from processor)
    master_adr_i            : IN std_logic_vector(ADDR_WIDTH - 1 DOWNTO 0);
    master_stb_i            : IN std_logic;   	-- bus cycle signal from "normal" bus master
    master_we_i             : IN std_logic;   	-- write enable from "normal" bus master
    rs232_rxd_i             : IN std_logic;   	-- Serial data from debug host terminal.
	 									-- Data bus (tri-state, to save interconnect)
    dat_i					: in std_logic_vector(DATA_WIDTH - 1 DOWNTO 0);   
	dat_o					: out std_logic_vector(DATA_WIDTH - 1 DOWNTO 0);   
    rst_o                   : OUT std_logic;   	-- Rst output to Wishbone "slaves"
    master_br_o             : OUT std_logic;   	-- Bus request to normal master device.
    stb_o                   : OUT std_logic;   	-- Bus cycle indicator to Wishbone "slaves"
    cyc_o                   : OUT std_logic;   	-- Bus cycle indicator to Wishbone "slaves"
	 									-- Address bus output to Wishbone "slaves"
    adr_o                   : OUT std_logic_vector(ADDR_WIDTH - 1 downto 0);  
    we_o                    : OUT std_logic;	-- Write enable to Wishbone "slaves"   
    rs232_txd_o             : OUT std_logic);   -- Serial transmit data to debug host terminal
end;

architecture rtl of rs232wb is

component uart_rx is
    Port (            serial_in : in std_logic;
                       data_out : out std_logic_vector(7 downto 0);
                    read_buffer : in std_logic;
                   reset_buffer : in std_logic;
                   en_16_x_baud : in std_logic;
            buffer_data_present : out std_logic;
                    buffer_full : out std_logic;
               buffer_half_full : out std_logic;
                            clk : in std_logic);
end component;

component uart_tx is
    Port (            data_in : in std_logic_vector(7 downto 0);
                 write_buffer : in std_logic;
                 reset_buffer : in std_logic;
                 en_16_x_baud : in std_logic;
                   serial_out : out std_logic;
                  buffer_full : out std_logic;
             buffer_half_full : out std_logic;
                          clk : in std_logic);
end component;

signal data_in, data_out : std_logic_vector(7 downto 0);
signal read_write : std_logic;
signal tx_buffer_full : std_logic;
signal buffer_data_present : std_logic;
signal write_buffer : std_logic;

signal count  : std_logic_vector(7 downto 0);
signal clk_div : std_logic;

-- 4-bit command, 16-bit address, 16-bit data 
--type cmd_type is array (8 downto 0) of std_logic_vector (3 downto 0);
signal cmd : std_logic_vector(31 downto 0);
signal cmd_rdy, rw : std_logic;

type state_type is (RESET,IDLE,WB,FIFO_LF1,FIFO_LF2,WB_1,
                    FIFO_D3,FIFO_D2,FIFO_D1,FIFO_D0,FIFO_CR,FIFO_PROMPT);
signal state, state_next : state_type;

type fifo_type is array (31 downto 0) of std_logic_vector(7 downto 0);
signal fifo : fifo_type;
signal fifo_write_ptr : std_logic_vector(4 downto 0);
signal fifo_read_ptr : std_logic_vector(4 downto 0);
signal write_fifo : std_logic;
signal fifo_in : std_logic_vector(7 downto 0);

signal data_bus : std_logic_vector(15 downto 0);

signal ascii_i : std_logic_vector(3 downto 0);
signal ascii_o : std_logic_vector(7 downto 0);

signal cap_data : std_logic;

begin

process (clk_i, rst_i)
begin
    if rst_i = '1' then
        state <= RESET;
        fifo_write_ptr <= (others=>'0');
    elsif rising_edge( clk_i ) then
        state <= state_next;
        if cap_data = '1' then
            data_bus <= dat_i;
        end if;
        if write_fifo = '1' then
            fifo(conv_integer(fifo_write_ptr)) <= fifo_in;
            fifo_write_ptr <= fifo_write_ptr + 1;
        end if;
    end if;
end process;

-- translate incoming data into ascii
process (ascii_i)
--    variable i : integer;
    variable tmp : std_logic_vector( 3 downto 0 );
begin
        if ascii_i < "1010" then
            ascii_o <= "0011" & ascii_i(3 downto 0);
        else
            tmp :=  ascii_i + "0111";
            ascii_o <= "0110" & tmp;
        end if;
end process;

-- translate serial data into wishbone commands
process (data_in, clk_i)
    variable valid : std_logic;
    variable vdata : std_logic_vector(3 downto 0);
    variable vnum : std_logic;
begin
    if rst_i = '1' then
        valid := '1';
    elsif rising_edge( clk_i ) then
        cmd_rdy <= '0';    
        if buffer_data_present = '1' and valid = '1' then
            valid := '0';
            vnum := '0';
            case data_in is
                when "00001000" => -- back space
                    cmd <= "0000" & cmd(31 downto 4);
                when "00001101" => cmd_rdy <= '1'; -- carriage return
                when "0011----" => -- data
                    vdata := data_in(3 downto 0);
                    vnum := '1';
                when "01-00001" =>  -- a or A
                    vdata := "1010";
                    vnum := '1';
                when "01-00010" =>  -- b or B
                    vdata := "1011";
                    vnum := '1';
                when "01-00011" =>  -- c or C
                    vdata := "1100";
                    vnum := '1';
                when "01-00100" =>  -- d or D
                    vdata := "1101";
                    vnum := '1';
                when "01-00101" =>  -- e or E
                    vdata := "1110";
                    vnum := '1';
                when "01-00110" =>  -- f or F
                    vdata := "1111";
                    vnum := '1';
                when "01-10010" =>  -- r of R
                    rw <= '0'; 
                when "01-10111" =>  -- w or W
                    rw <= '1';
                when others => null;
            end case;
            if vnum = '1' then
                cmd <= cmd(31-4 downto 0) & vdata;
            end if;
        elsif buffer_data_present = '0' then 
            valid := '1';  -- the buffer_data_present is slow to be cleared, so wait ...
        end if;
    end if;
end process;


process (state, ack_i, cmd_rdy, data_bus, ascii_o)
begin
    -- default values
    cap_data <= '0';
    we_o <= '0';
    stb_o <= '0';
    cyc_o <= '0';
    write_fifo <= '0';

    case state is
    when RESET =>
        state_next <= IDLE;
        write_fifo <= '0';
    when IDLE =>
        if cmd_rdy = '1' then
            state_next <= WB;
        else
            state_next <= IDLE;
        end if;
    when WB_1 =>
        cap_data <= '1';
        state_next <= FIFO_LF1;
    when WB =>
        we_o <= rw;
        stb_o <= '1';
        cyc_o <= '1';
        dat_o <= cmd(15 downto 0);
        if rw = '0' then
            adr_o <= cmd(15 downto 0);
        else
            adr_o <= cmd(31 downto 16);
        end if;
        if ack_i = '1' then
            state_next <= WB_1;
        else
            state_next <= WB;
        end if;
    when FIFO_LF1 =>
        write_fifo <= '1';        
        fifo_in <= "00001010"; -- line feed
        if rw = '1' then
            state_next <= FIFO_PROMPT;
        else
            state_next <= FIFO_D3;
        end if;
    when FIFO_D3 =>
        write_fifo <= '1';        
        ascii_i <= data_bus(15 downto 12);
        fifo_in <= ascii_o;
        state_next <= FIFO_D2;
    when FIFO_D2 =>
        write_fifo <= '1';        
        ascii_i <= data_bus(11 downto 8);
        fifo_in <= ascii_o;
        state_next <= FIFO_D1;
    when FIFO_D1 =>
        write_fifo <= '1';        
        ascii_i <= data_bus(7 downto 4);
        fifo_in <= ascii_o;
        state_next <= FIFO_D0;
    when FIFO_D0 =>
        write_fifo <= '1';        
        ascii_i <= data_bus(3 downto 0);
        fifo_in <= ascii_o;
        state_next <= FIFO_LF2;
    when FIFO_LF2 =>
        write_fifo <= '1';        
        fifo_in <= "00001010"; -- line feed
        state_next <= FIFO_CR;
    when FIFO_CR =>
        write_fifo <= '1';        
        fifo_in <= "00001101"; -- CR
        state_next <= FIFO_PROMPT;
    when FIFO_PROMPT =>  
        write_fifo <= '1';        
        fifo_in <= "00111110"; -- >
        state_next <= IDLE;
    when others =>
        state_next <= RESET;
    end case;
end process;
        
----------------------------------------------------------------------
-- RS232 components
----------------------------------------------------------------------

-- rs232 clock divider
process (clk_i, rst_i)
begin
	if rising_edge(clk_i) then
		count <= count + 1;
		if count = "10100100" then
			count <= (others=>'0');
			clk_div <= not clk_div;
        end if;
	end if;
end process;

process (clk_div, rst_i)
begin
    if rst_i = '1' then
        fifo_read_ptr <= (others=>'0');    
    elsif rising_edge( clk_div ) then 
        write_buffer <= '0';
        if buffer_data_present = '0' then
            if fifo_write_ptr /= fifo_read_ptr then
                data_out <= fifo(conv_integer(fifo_read_ptr));
                fifo_read_ptr <= fifo_read_ptr + 1;
                write_buffer <= '1';
            end if;
        elsif buffer_data_present = '1' then
            data_out <= data_in;
            write_buffer <= '1';
        end if;
    end if;
end process;
       

uart_rx_inst : uart_rx 
    port map(            serial_in => rs232_rxd_i,
                       data_out => data_in,
                    read_buffer => '1',
                   reset_buffer => rst_i,
                   en_16_x_baud => '1',
            buffer_data_present => buffer_data_present,
--                    buffer_full => rx_buffer_full,
--               buffer_half_full => null,
                            clk => clk_div
);

uart_tx_inst : uart_tx
    port map (            data_in => data_out,
                 write_buffer => write_buffer,
                 reset_buffer => rst_i,
                 en_16_x_baud => '1',
                   serial_out => rs232_txd_o,
                  buffer_full => tx_buffer_full,
--             buffer_half_full => null,
                          clk	=> clk_div
	);

end rtl;
