library IEEE;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

library UNISIM;
use UNISIM.VComponents.all;

entity toplevel is
   PORT (
        clk             : IN  std_logic;  -- System clock input   
        rst             : IN  std_logic;  -- Reset signal for this module   
        rs232_rxd_i     : IN  std_logic;  -- Serial data from debug host terminal.   
        rs232_txd_o     : OUT std_logic;  -- Serial transmit data to debug host terminal   

        led             : out std_logic;  -- LED1 on the D2 board

		-- outputs to/from the D2 IO board
        IO_clk : out std_logic;									--output clock (divided)
		IO_cs : out std_logic;										--chip select
        IO_we : out std_logic;										--write enable
        IO_oe : out std_logic;										--output enable
 		IO_addr : out std_logic_vector(5 downto 0);			--address bus for communications with CPLD
        IO_data : inout std_logic_vector(7 downto 0)			--bidir data bus between CPLD & design
	);
end;

architecture rtl of toplevel is

constant DATA_WIDTH : integer := 16;
constant ADDR_WIDTH : integer := 16;

component dpblockram is 
 port (
 	-- wishbone signals
   clk_i  : in std_logic; 
	stb_i: in std_logic;
 	we_i : in std_logic; 
 	adr_i: in std_logic_vector(4 downto 0); 
 	dat_i: in std_logic_vector(15 downto 0); 
	ack_o: out std_logic;
 	dat_o: out std_logic_vector(15 downto 0);
   -- non wishbone signals
	dpra : in std_logic_vector(4 downto 0); 
 	dpo  : out std_logic_vector(15 downto 0)); 
 end component; 

 component wbd2io is
    Port ( 
	 		  -- wishbone signals
        clk_i : in std_logic;
        stb_i : in std_logic;
        we_i  : in std_logic;
		adr_i : in std_logic_vector(3 downto 0);
        dat_i : in std_logic_vector(15 downto 0);
        dat_o : out std_logic_vector(15 downto 0);
        ack_o : out std_logic;

			  -- outputs to/from the D2 IO board
		IO_clk : out std_logic;									--output clock (divided)
		cs : out std_logic;										--chip select
        we : out std_logic;										--write enable
        oe : out std_logic;										--output enable
 		addr : out std_logic_vector(5 downto 0);			--address bus for communications with CPLD
        data : inout std_logic_vector(7 downto 0)			--bidir data bus between CPLD & design
    );
end component;

component rs232wb is
    port (
    clk_i                   : IN std_logic;	-- System clock input			   
    rst_i                 : IN std_logic;   	-- Reset signal for this module
    ack_i                   : IN std_logic;   	-- Ack input from Wishbone "slaves"
    err_i                   : IN std_logic;   	-- Err input from Wishbone "slaves"
    master_bg_i             : IN std_logic;   	-- Bus Grant (grants this module the bus)
                             					-- Address bus input from "normal" Wishbone
                             					-- master (i.e. from processor)
    master_adr_i            : IN std_logic_vector(ADDR_WIDTH - 1 DOWNTO 0);
    master_stb_i            : IN std_logic;   	-- bus cycle signal from "normal" bus master
    master_we_i             : IN std_logic;   	-- write enable from "normal" bus master
    rs232_rxd_i             : IN std_logic;   	-- Serial data from debug host terminal.
	 									-- Data bus (tri-state, to save interconnect)
    dat_i					: in std_logic_vector(DATA_WIDTH - 1 DOWNTO 0);   
	dat_o					: out std_logic_vector(DATA_WIDTH - 1 DOWNTO 0);   
    rst_o                   : OUT std_logic;   	-- Rst output to Wishbone "slaves"
    master_br_o             : OUT std_logic;   	-- Bus request to normal master device.
    stb_o                   : OUT std_logic;   	-- Bus cycle indicator to Wishbone "slaves"
    cyc_o                   : OUT std_logic;   	-- Bus cycle indicator to Wishbone "slaves"
	 									-- Address bus output to Wishbone "slaves"
    adr_o                   : OUT std_logic_vector(ADDR_WIDTH - 1 downto 0);  
    we_o                    : OUT std_logic;	-- Write enable to Wishbone "slaves"   
    rs232_txd_o             : OUT std_logic);   -- Serial transmit data to debug host terminal
end component;



   SIGNAL ack_i                    :  std_logic;	-- Ack input from Wishbone "slaves"   
   SIGNAL err_i                    :  std_logic;	-- Err input from Wishbone "slaves"   
   SIGNAL master_bg_i              :  std_logic;	-- Bus Grant (grants this module the bus)   
   
   										-- Address bus input from "normal" Wishbone
										-- master (i.e. from processor)
   SIGNAL master_adr_i             :  std_logic_vector(ADDR_WIDTH - 1 DOWNTO 0);   
   SIGNAL master_stb_i             :  std_logic;	-- bus cycle signal from "normal" bus master   
   SIGNAL master_we_i              :  std_logic;	-- write enable from "normal" bus master   

  										-- Data bus (tri-state, to save interconnect)

    signal dat_i						  :  std_logic_vector(DATA_WIDTH - 1 DOWNTO 0);   
	signal dat_o						  :  std_logic_vector(DATA_WIDTH - 1 DOWNTO 0);   
    signal rst_o                    :  std_logic;	-- Rst output to Wishbone "slaves"   
   SIGNAL master_br_o              :  std_logic;	-- Bus request to normal master device.   
   SIGNAL stb_o                    :  std_logic;	-- Bus cycle indicator to Wishbone "slaves"   
   SIGNAL cyc_o                    :  std_logic;	-- Bus cycle indicator to Wishbone "slaves"   

   										-- Address bus output to Wishbone "slaves"
   SIGNAL adr_o                    :  std_logic_vector(15 DOWNTO 0);   
   SIGNAL we_o                     :  std_logic;	-- Write enable to Wishbone "slaves"   
	   

	signal dat_o_dp, dat_o_io, dpo  : std_logic_vector(15 downto 0);
	signal adr_nocon				: std_logic_vector(4 DOWNTO 0); 

	signal act_dp, act_io				: std_logic;
	signal ack_o_dp, ack_o_io			: std_logic;
	signal stb_i_io, stb_i_dp			: std_logic;

    signal rst_i : std_logic;
BEGIN

ibufg_rst : IBUFG port map( rst_i, rst);

process( rst_i )
begin
    if rst_i = '1' then
        led <= '1';
    else
        led <= '0';
    end if;
end process;


rs232wb_inst : rs232wb
    port map (
        clk_i => clk,
        rst_i => rst_i,
        ack_i => ack_i,
        err_i => '0', --err_i,
        master_bg_i => '1', --master_bg_i,
        master_adr_i => master_adr_i,
        master_stb_i => '0', --master_stb_i,
        master_we_i => '0', --master_we_i,
        rs232_rxd_i => rs232_rxd_i,
		dat_o => dat_o,
		dat_i => dat_i,
        rst_o => rst_o,
        master_br_o => master_br_o,
        stb_o => stb_o,
        cyc_o => cyc_o,
        adr_o => adr_o,
        we_o => we_o,
        rs232_txd_o => rs232_txd_o
); 
  

dualport_unit : dpblockram
	port map (
		clk_i  => clk,
		stb_i  => stb_i_dp,
		we_i   => we_o, 
		adr_i  => adr_o(4 downto 0), 
		dat_i  => dat_o,
		ack_o  => ack_o_dp,
		dat_o  => dat_o_dp,

		dpra => adr_nocon, 
		dpo => dpo
    ); 

wbd2io_unit : wbd2io
    port map ( 
	    -- wishbone signals
	 	clk_i => clk, 
        stb_i => stb_i_io, 
        we_i  => we_o, 
		adr_i => adr_o(3 downto 0),
        dat_i => dat_o,
        dat_o => dat_o_io,
        ack_o => ack_o_io,

		-- outputs to/from the D2 IO board
		IO_clk => IO_clk,	--output clock (divided)
		cs => IO_cs,	--chip select
        we => IO_we, 	--write enable
        oe => IO_oe, 	--output enable
 		addr => IO_addr,	--address bus for communications with CPLD
        data => IO_data	--bidir data bus between CPLD & design
	);
	
   
	-- based on the address determine which slave will be activated
    arbiter : process (clk, adr_o)
    begin
		act_dp <= '0';
		act_io <= '0';
		if ( adr_o(4) = '0' ) then
			act_dp <= '1';
		else
			act_io <= '1';
		end if;
	end process;

	-- combine the master strobe (stb_o) signal with the address
	-- decoded enables (act_xx);
	stb_i_io <= stb_o and cyc_o and act_io;
	stb_i_dp <= stb_o and cyc_o and act_dp;

	-- select the data ouput of the slave who is active
	dat_i <= dat_o_dp when (act_dp = '1') else (others => 'Z');
	dat_i <= dat_o_io when (act_io = '1') else (others => 'Z');

	-- OR all the slave acknolodge signals together and pass it to the master
	ack_i <= ack_o_dp or ack_o_io;
	

end rtl;
